@@ -29,6 +29,7 @@ from __future__ import absolute_import
 import os
 import shutil
 import tempfile
+import threading
 import warnings
 import weakref
 
@@ -37,6 +38,9 @@ import cffi
 __all__ = ['Environment', 'Cursor', 'Transaction', 'open', 'Error',
            'enable_drop_gil']
 
+# Used to track context across cffi callbcks.
+_callbacks = threading.local()
+
 _ffi = cffi.FFI()
 _ffi.cdef('''
     typedef int mode_t;
@@ -105,6 +109,7 @@ _ffi.cdef('''
     int mdb_env_open(MDB_env *env, const char *path, unsigned int flags,
                      mode_t mode);
     int mdb_env_copy(MDB_env *env, const char *path);
+    int mdb_env_copyfd(MDB_env *env, int fd);
     int mdb_env_stat(MDB_env *env, MDB_stat *stat);
     int mdb_env_info(MDB_env *env, MDB_envinfo *stat);
     int mdb_env_sync(MDB_env *env, int force);
@@ -131,6 +136,10 @@ _ffi.cdef('''
     int mdb_cursor_count(MDB_cursor *cursor, size_t *countp);
     int mdb_cursor_get(MDB_cursor *cursor, MDB_val *key, MDB_val*data, int op);
 
+    typedef int (MDB_msg_func)(const char *msg, void *ctx);
+    int mdb_reader_list(MDB_env *env, MDB_msg_func *func, void *ctx);
+    int mdb_reader_check(MDB_env *env, int *dead);
+
     #define EINVAL ...
     #define MDB_APPEND ...
     #define MDB_CREATE ...
@@ -317,6 +326,13 @@ def enable_drop_gil():
     *Caution:* this function should be invoked before any threads are created.
     """
 
+@_ffi.callback("int(char *, void *)")
+def _msg_func(s, _):
+    """mdb_msg_func() callback. Appends `s` to _callbacks.msg_func list.
+    """
+    _callbacks.msg_func.append(_ffi.string(s))
+    return 0
+
 
 class Environment(object):
     """
@@ -500,6 +516,17 @@ class Environment(object):
         if rc:
             raise Error("mdb_env_copy", rc)
 
+    def copyfd(self, fd):
+        """Copy a consistent version of the environment to file descriptor
+        `fd`.
+
+        Equivalent to `mdb_env_copyfd()
+        <http://symas.com/mdb/doc/group__mdb.html#ga5d51d6130325f7353db0955dbedbc378>`_
+        """
+        rc = mdb_env_copyfd(self._env, fd)
+        if rc:
+            raise Error("mdb_env_copyfd", rc)
+
     def sync(self, force=False):
         """Flush the data buffers to disk.
 
@@ -592,6 +619,29 @@ class Environment(object):
             "num_readers": info.me_numreaders
         }
 
+    def readers(self):
+        """Return a list of newline-terminated human readable strings
+        describing the current state of the reader lock table.
+        """
+        _callbacks.msg_func = []
+        try:
+            rc = mdb_reader_list(self._env, _msg_func, _ffi.NULL)
+            if rc:
+                raise Error("mdb_reader_list", rc)
+            return _callbacks.msg_func
+        finally:
+            del _callbacks.msg_func
+
+    def check_readers(self):
+        """Search the reader lock table for stale entries, for example due to a
+        crashed process. Returns the number of stale entries that were cleared.
+        """
+        reaped = _ffi.new('int[]', 1)
+        rc = mdb_reader_check(self._env, reaped)
+        if rc:
+            raise Error('mdb_reader_check', rc)
+        return reaped[0]
+
     def open_db(self, name=None, txn=None, reverse_key=False, dupsort=False,
             create=True):
         """
@@ -924,7 +974,7 @@ class Transaction(object):
     def put(self, key, value, dupdata=False, overwrite=True, append=False,
             db=None):
         """Store a record, returning ``True`` if it was written, or ``False``
-        to indicate the key was already present and `override=False`.
+        to indicate the key was already present and `overwrite=False`.
 
         Equivalent to `mdb_put()
         <http://symas.com/mdb/doc/group__mdb.html#ga4fa8573d9236d54687c61827ebf8cac0>`_
